
// Copyright University of Warwick 2016 - 2017.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

// Authors:
// Thomas Latham
// John Back
// Paul Harrison

/*! \file LauRooFitSlave.hh
    \brief File containing declaration of LauRooFitSlave class.
*/

/*! \class LauRooFitSlave
    \brief A class for creating a RooFit-based slave process for simultaneous/combined fits

    Implementation of the JFit method described in arXiv:1409.5080 [physics.data-an].
*/

#ifndef LAU_ROO_FIT_SLAVE
#define LAU_ROO_FIT_SLAVE

#include <utility>
#include <vector>

#include "RooAbsPdf.h"
#include "RooAbsData.h"
#include "RooCategory.h"
#include "RooNLLVar.h"
#include "TMatrixDfwd.h"
#include "TString.h"

#include "LauSimFitSlave.hh"

class LauParameter;


class LauRooFitSlave : public LauSimFitSlave {

	public:
		//! Constructor
		LauRooFitSlave( RooAbsPdf& model, const Bool_t extended, const RooArgSet& vars, const TString& weightVarName = "" );

		//! Destructor
		virtual ~LauRooFitSlave();

		//! Initialise the fit model
		virtual void initialise();

		//! This function sets the parameter values from Minuit
		/*! 
			\param [in] par an array storing the various parameter values
			\param [in] npar the number of free parameters
		*/
		virtual void setParsFromMinuit(Double_t* par, Int_t npar);

		//! Calculates the total negative log-likelihood
		virtual Double_t getTotNegLogLikelihood();

	protected:

		//! Package the initial fit parameters for transmission to the master
		/*!
		    \param [out] array the array to be filled with the LauParameter objects
		*/	
		virtual void prepareInitialParArray( TObjArray& array );

		//! Convert a RooRealVar into a LauParameter
		/*!
		    \param [in] rooParameter the RooRealVar to be converted
		    \return the newly created LauParameter
		*/
		LauParameter* convertToLauParameter( const RooRealVar* rooParameter ) const;

		//! Convert a RooFormulaVar into LauParameters
		/*!
		    \param [in] rooFormula the RooFormulaVar to be converted
		    \return the vector of pointers to the RooRealVars and newly created LauParameters
		*/
		std::vector< std::pair<RooRealVar*,LauParameter*> > convertToLauParameters( const RooFormulaVar* rooFormula ) const;

		//! Perform all finalisation actions
		/*!
			- Receive the results of the fit from the master
			- Perform any finalisation routines
		  	- Package the finalised fit parameters for transmission back to the master

			\param [in] fitStat the status of the fit, e.g. status code, EDM, NLL
			\param [in] parsFromMaster the parameters at the fit minimum
			\param [in] covMat the fit covariance matrix
			\param [out] parsToMaster the array to be filled with the finalised LauParameter objects
		*/	
		virtual void finaliseExperiment( const LauAbsFitter::FitStatus& fitStat, const TObjArray* parsFromMaster, const TMatrixD* covMat, TObjArray& parsToMaster );

		//! Open the input file and verify that all required variables are present
		/*!
			\param [in] dataFileName the name of the input file
			\param [in] dataTreeName the name of the input tree
		*/	
		virtual Bool_t verifyFitData(const TString& dataFileName, const TString& dataTreeName);

		//! Read in the data for the current experiment
		/*!
			\return the number of events read in
		*/	
		virtual UInt_t readExperimentData();

		//! Cache the input data values to calculate the likelihood during the fit
		virtual void cacheInputFitVars();

	private:
		//! Cleanup the data
		void cleanData();

		//! Copy constructor (not implemented)
		LauRooFitSlave(const LauRooFitSlave& rhs);

		//! Copy assignment operator (not implemented)
		LauRooFitSlave& operator=(const LauRooFitSlave& rhs);

		//! The model
		RooAbsPdf& model_;

		//! The dataset variables
		RooArgSet dataVars_;

		//! The name of the (optional) weight variable in the dataset
		TString weightVarName_;

		//! The data file
		TFile* dataFile_;

		//! The data tree
		TTree* dataTree_;

		//! The data for the current experiment
		RooAbsData* exptData_;

		//! Is the PDF extended?
		const Bool_t extended_;

		//! The experiment category variable
		RooCategory iExptCat_;

		//! The NLL variable
		RooNLLVar* nllVar_;

		//! The fit parameters (as RooRealVar's)
		std::vector<RooRealVar*> fitVars_;

		//! The fit parameters (as LauParameter's)
		std::vector<LauParameter*> fitPars_;

		ClassDef(LauRooFitSlave,0);
};

#endif
